import typing
import inspect
import logging
import json
from dataclasses import MISSING, asdict
from abc import ABC, abstractmethod

if typing.TYPE_CHECKING:
    from .base import BaseModel


logger = logging.getLogger(__name__)


def get_function_call_args(
    func, params: typing.Union[typing.Dict[str, typing.Any], object]
) -> typing.Dict[str, typing.Any]:
    """
    Extracts the arguments for a function call from the provided parameters.

    Args:
        func: The function for which arguments are to be extracted.
        params: A dictionary of parameters containing
                the necessary arguments for the function.

    Returns:
        A dictionary where the keys are the function argument names
        and the values are the corresponding argument values.
    """
    params_dict = {}
    try:
        sig = inspect.signature(func)
        for param in sig.parameters.values():
            if param.name != "self":
                value = (
                    params.get(param.name, param.default)
                    if isinstance(params, dict)
                    else getattr(params, param.name, param.default)
                )
                if value is not MISSING and value is not inspect.Parameter.empty:
                    params_dict[param.name] = value
                else:
                    params_dict[param.name] = None
    except (ValueError, KeyError) as e:
        logger.warning(f"Parsing {func} for call parameters failed {str(e)}")

    for key in ["args", "kwargs"]:
        if key in params_dict and params_dict[key] is None:
            params_dict.pop(key, None)
    return params_dict


class BaseModelFormatter(ABC):
    format_name: str = None

    @abstractmethod
    def encode(
        self, _type: typing.Type["BaseModel"], obj: typing.Dict[str, typing.Any]
    ) -> "BaseModel":
        pass

    @abstractmethod
    def decode(self, instance: "BaseModel") -> typing.Any:
        pass

    @classmethod
    def get_formatters(cls):
        for subclass in cls.__subclasses__():
            yield from subclass.get_formatters()
            yield subclass

    @classmethod
    def get_formatter(cls, *args, format_name: str, **kwargs) -> "BaseModelFormatter":
        for subclass in cls.get_formatters():
            if subclass.format_name == format_name:
                return subclass(*args, **kwargs)
        raise KeyError(f"Format {format_name} not found")


class DictModelFormatter(BaseModelFormatter):
    format_name = "dict"

    def _encode(
        self, _type: typing.Type["BaseModel"], obj: typing.Dict[str, typing.Any]
    ) -> "BaseModel":
        kwargs = get_function_call_args(_type.__init__, obj)
        excluded_kwargs = {
            key: obj[key] for key in obj.keys() if key not in kwargs.keys()
        }
        instance = _type(**kwargs)
        if excluded_kwargs:
            instance.__dict__.update(excluded_kwargs)
        # force execute post init again for normal field validation
        instance.__post_init__()
        return instance

    def encode(
        self,
        _type: typing.Type["BaseModel"],
        obj: typing.Union[
            typing.Dict[str, typing.Any], typing.List[typing.Dict[str, typing.Any]]
        ],
    ) -> typing.Union["BaseModel", typing.List["BaseModel"]]:
        if isinstance(obj, dict):
            return self._encode(_type, obj)
        elif isinstance(obj, list):
            content = []
            for item in obj:
                content.append(self._encode(_type, item))
            return content
        else:
            raise TypeError("Object must be dict or list")

    def decode(self, instance: "BaseModel") -> typing.Dict[str, typing.Any]:
        return asdict(instance)


class JSONModelFormatter(DictModelFormatter):
    format_name = "json"

    def encode(
        self, _type: typing.Type["BaseModel"], obj: str
    ) -> typing.Union["BaseModel", typing.List["BaseModel"]]:
        obj = json.loads(obj)
        if isinstance(obj, dict):
            return super().encode(_type, obj)
        elif isinstance(obj, list):
            content = []
            for value in obj:
                content.append(super().encode(_type, value))
            return content
        else:
            raise TypeError(f"Type {obj} is not JSON serializable")

    def decode(self, instance: "BaseModel") -> str:
        return json.dumps(super().decode(instance))



import unittest
import typing
from dataclasses import field
from pydantic_mini import BaseModel, MiniAnnotated, Attrib
from pydantic_mini.exceptions import ValidationError


class TestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        class MyModel(BaseModel):
            name: str
            age: int

        class DataClassField(BaseModel):
            school = field(default="knust")
            value = field(default_factory=lambda: 1)

        class AnnotatedDataClass(BaseModel):
            email: MiniAnnotated[
                str, Attrib(pattern=r"^[^@]+@[^@]+\.[^@]+$", max_length=13)  # noqa:
            ]
            value: MiniAnnotated[int, Attrib(gt=4, lt=20, default=5)]

        class UsingOptionalDataClass(BaseModel):
            value: typing.Optional[int]
            name: MiniAnnotated[typing.Optional[str], Attrib(max_length=20)]

        cls.MyModel = MyModel
        cls.DataClassField = DataClassField
        cls.AnnotatedDataClass = AnnotatedDataClass
        cls.UsingOptionalDataClass = UsingOptionalDataClass

    def test_simple_annotated_model(self):
        instance = self.MyModel(name="test", age=10)
        self.assertEqual(instance.name, "test")
        self.assertEqual(instance.age, 10)

        with self.assertRaises(TypeError):
            self.MyModel(name=12, age="hello")

    def test_dataclass_field(self):
        instance = self.DataClassField()
        self.assertEqual(instance.school, "knust")
        self.assertEqual(instance.value, 1)

        # validate detected type from default
        with self.assertRaises(TypeError):
            self.DataClassField(school=23, value="hello")

    def test_mini_annotated_annotation(self):
        instance = self.AnnotatedDataClass(value=10, email="ex@email.com")
        self.assertEqual(instance.email, "ex@email.com")
        self.assertEqual(instance.value, 10)

        with self.assertRaises(ValidationError):
            self.AnnotatedDataClass(value=10, email="ex")

        with self.assertRaises(ValidationError):
            self.AnnotatedDataClass(value=10, email="looooooooong-email@example.com")

    def test_fields_with_or_without_default_values_cause_error(self):
        class Person(BaseModel):
            name: str = "nafiu"
            school: str

        class Person1(BaseModel):
            name: str
            school: str = "knust"

        p1 = Person(school="knust")
        self.assertEqual(p1.name, "nafiu")
        self.assertEqual(p1.school, "knust")

        p2 = Person1(name="nafiu")
        self.assertEqual(p2.name, "nafiu")
        self.assertEqual(p2.school, "knust")

        # validate positional arguments are required
        with self.assertRaises(TypeError):
            Person(name="nafiu")

        with self.assertRaises(TypeError):
            Person1(school="knust")

    def test_figured_out_optional_field_from_annotation_has_none_value(self):
        p = self.UsingOptionalDataClass()
        self.assertEqual(p.value, None)
        self.assertEqual(p.name, None)

    def test_model_creation_with_dict(self):
        param = {"name": "nafiu", "age": 12}
        instance = self.MyModel.loads(param, _format="dict")
        self.assertEqual(instance.name, "nafiu")
        self.assertEqual(instance.age, 12)

    def test_model_serialization_with_dict(self):
        instance = self.MyModel(name="nafiu", age=12)
        _dict = instance.dump(_format="dict")
        self.assertIsInstance(_dict, dict)
        self.assertEqual(_dict, {"name": "nafiu", "age": 12})

    def test_multiple_model_creation_with_dict(self):
        params = [
            {"name": "nafiu", "age": 12},
            {"name": "shaibu", "age": 13},
            {"name": "nshaibu", "age": 14},
        ]
        instance = self.MyModel.loads(params, _format="dict")
        self.assertIsInstance(instance, list)
        self.assertEqual(len(instance), len(params))
